from math import prod
from numbers import Number
from typing import Literal

from torch import Tensor, cat, nn, randn


class VisionTransformer(nn.Module):
    """Configurable vision transformer (ViT).

    Implements the vision transformer as proposed by Dosovitskiy et al., ICLR 2021.

    Args:
        num_channels (int): Number of channels of the input.
        input_sizes (list of int): Spatial sizes of the input.
        output_size (int): Size of the output layer.
        patch_size (int): Size of a patch in one dimension.
        num_heads (int): Number of attention heads in each encoder block.
        latent_size_multiplier (int): Yields the size of the embedding when multiplied
            with `num_heads`.
        num_layers (int): Number of encoder blocks.
        encoder_size (int): Number of hidden units in each encoder MLP.
        head_size (int or list of int): Sizes of hidden layers in MLP head.
        dropout (float, optional): Dropout probabilities of embedding, encoder and MLP
            head.  Default: 0.
        encoder_activation (str, optional): Encoder activation function string, either
            `'relu'` or `'gelu'`.  Default: `'gelu'`.
        head_activation (str, optional): MLP head activation function string, `'relu'`,
            `'gelu'` or `'tanh'`.  Default: `'gelu'`.
    """

    def __init__(
        self,
        num_channels: int,
        input_sizes: list[int],
        output_size: int,
        patch_size: int,
        num_heads: int,
        latent_size_multiplier: int,
        num_layers: int,
        encoder_size: int,
        head_size: int | list[int],
        dropout: float = 0,
        encoder_activation: Literal["relu", "gelu"] = "gelu",
        head_activation: Literal["relu", "gelu", "tanh"] = "gelu",
    ) -> None:
        super().__init__()
        self.kwargs = {
            "num_channels": num_channels,
            "input_sizes": input_sizes,
            "output_size": output_size,
            "patch_size": patch_size,
            "num_heads": num_heads,
            "latent_size_multiplier": latent_size_multiplier,
            "num_layers": num_layers,
            "encoder_size": encoder_size,
            "head_size": head_size,
            "dropout": dropout,
            "encoder_activation": encoder_activation,
            "head_activation": head_activation,
        }
        latent_size = latent_size_multiplier * num_heads
        self.embedding = Embedding(
            num_channels, input_sizes, patch_size, latent_size, dropout
        )
        layer_norm = nn.LayerNorm(latent_size)
        encoder_layer = nn.TransformerEncoderLayer(
            d_model=latent_size,
            nhead=num_heads,
            dim_feedforward=encoder_size,
            dropout=dropout,
            activation=encoder_activation,
            batch_first=True,
            norm_first=True,
        )
        self.encoder = nn.TransformerEncoder(
            encoder_layer, num_layers, layer_norm, enable_nested_tensor=False
        )
        if isinstance(head_size, Number):
            head_size = [head_size]
        self.mlp_head = MLP(
            input_size=latent_size,
            output_size=output_size,
            hidden_sizes=head_size,
            dropout=dropout,
            activation=head_activation,
        )

    def forward(self, data: Tensor) -> Tensor:
        data = self.embedding(data)
        data = self.encoder(data)
        # Take encoder output corresponding to class_token
        output = self.mlp_head(data[:, 0, :])
        return output


class Embedding(nn.Module):
    """An embedding for a vision transformer.

    Splits the input into patches, and projects the patches.  Also adds a class token
    and a position embedding.

    Args:
        num_channels (int): Number of channels of the input.
        input_sizes (list of int): Spatial sizes of the input.
        patch_size (int): Size of a patch in one dimension.
        latent_size (int): Size of the embedding.
        dropout (float, optional): Dropout probability.  Default: 0.
    """

    def __init__(
        self,
        num_channels: int,
        input_sizes: list[int],
        patch_size: int,
        latent_size: int,
        dropout: float = 0,
    ) -> None:
        super().__init__()
        # Use Unfold to split the input image into patches
        self.unfold = nn.Unfold(kernel_size=patch_size, stride=patch_size)
        num_patches = prod(
            [(input_size - patch_size) // patch_size + 1 for input_size in input_sizes]
        )
        flattened_size = num_channels * patch_size ** len(input_sizes)
        self.linear = nn.Linear(flattened_size, latent_size)
        self.class_token = nn.Parameter(randn(1, 1, latent_size))
        self.position_embeddings = nn.Parameter(randn(1, num_patches + 1, latent_size))
        self.dropout = nn.Dropout(dropout)

    def forward(self, data: Tensor) -> Tensor:
        # Split images into patches and flatten into
        # batch_size x num_patches x (num_channels * patch_size ** len(input_sizes))
        batch_size = data.shape[0]
        data = self.unfold(data).permute(0, 2, 1).contiguous()
        data = self.linear(data)
        output = self.dropout(
            cat([self.class_token.expand(batch_size, -1, -1), data], dim=1)
            + self.position_embeddings
        )
        return output


class MLP(nn.Module):
    """Configurable multilayer perceptron.

    Args:
        input_size (int): Size of the input layer.
        output_size (int): Size of the output layer.
        hidden_sizes (list of int, optional): Sizes of hidden layers.  Default: `None`.
        dropout (float or list of float, optional): Dropout probabilities of each
            hidden layer.  If `None`, no dropout will be used.  If single float, the
            same dropout probability will be used for all hidden layers.
            Default: `None`.
        activation (str, optional): Activation function string, `'relu'`, `'gelu'` or
            `'tanh'`.  Default: `'relu'`.
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        hidden_sizes: list[int] = None,
        dropout: float | list[float] = None,
        activation: Literal["relu", "gelu", "tanh"] = "relu",
    ) -> None:
        super().__init__()
        self.flatten = nn.Flatten()
        if hidden_sizes is None:
            hidden_sizes = []
        layers = [
            nn.Linear(i, j)
            for i, j in zip([input_size] + hidden_sizes, hidden_sizes + [output_size])
        ]
        if dropout is None:
            dropout_modules = []
        elif isinstance(dropout, Number):
            dropout_modules = [nn.Dropout(dropout) for _ in range(len(hidden_sizes))]
        else:
            dropout_modules = [nn.Dropout(rate) for rate in dropout]
        modules = []
        for i, layer in enumerate(layers):
            modules.append(layer)
            if i < len(layers) - 1:
                if activation == "relu":
                    modules.append(nn.ReLU())
                elif activation == "gelu":
                    modules.append(nn.GELU())
                elif activation == "tanh":
                    modules.append(nn.Tanh())
                else:
                    raise ValueError(f"unknown activation '{activation}'")
                if len(dropout_modules) > i:
                    modules.append(dropout_modules[i])
        self.linear_stack = nn.Sequential(*modules)

    def forward(self, data: Tensor) -> Tensor:
        data = self.flatten(data)
        output = self.linear_stack(data)
        return output



import argparse
from typing import Any, Callable

import torch

from .model import VisionTransformer
from .predict import prediction_loss
from .preprocess import get_train_loaders_mnist
from .utils import FILE_LIKE, save_model


def train_mnist(
    config: dict[str, int | float | list[int]],
    data_dir: str,
    use_validation: bool = True,
    use_augmentation: bool = True,
    model_file: FILE_LIKE = None,
    report_fn: Callable[
        [
            int,
            float,
            float,
            torch.nn.Module,
            torch.optim.Optimizer,
            torch.optim.lr_scheduler.LRScheduler,
        ],
        None,
    ] = None,
    resume_states: dict[str, int | dict[str, Any]] = None,
    device: torch.device = "cpu",
) -> None:
    """Trains a single model on MNIST and eventually saves the model.

    Args:
        config (dict): Training configuration including `'batch_size'`, `'num_epochs'`,
            `'lr'`, `'weight_decay'`, `'epoch_lr_restart'`, `'patch_size'`,
            `'num_heads'`, `'latent_size_multiplier'`, `'num_layers'`, `'encoder_size'`,
            `'head_size'` and `'dropout'`.
        data_dir (str): Directory of the MNIST dataset.
        use_validation (bool, optional): If true, sets aside a validation set from the
            training set, else uses all training samples for training.  Default: `True`.
        use_augmentation (bool, optional): If true, augments the training dataset with
            random affine transformations.  Default: `True`.
        model_file (FILE_LIKE, optional): File name to save the model to.  If `None`
            then the model is not saved.  Default: `None`.
        report_fn (callable, optional): A function for reporting the training state.
            The function must accept arguments for epoch number (`int`),
            training loss (`float`), validation loss (`float`),
            model (`torch.nn.Module`),
            optimizer (`torch.optim.Optimizer`) and
            lr_scheduler (`torch.optim.lr_scheduler.LRScheduler`).  Default: `None`.
        resume_states (dict, optional): Dictionary with states for `'epoch'`, `'model'`,
            `'optimizer'` and `'lr_scheduler'`.  Default: `None`.
        device (torch.device, optional): Device to train the model on.
            Default: `'cpu'`.
    """
    train_fraction = 0.8 if use_validation else 1.0
    train_loader, val_loader = get_train_loaders_mnist(
        data_dir=data_dir,
        batch_size=config["batch_size"],
        train_fraction=train_fraction,
        use_augmentation=use_augmentation,
    )
    model = init_mnist_model(config, device)
    if resume_states is not None:
        model.load_state_dict(resume_states["model"])
    loss_fn = torch.nn.CrossEntropyLoss()
    train(
        model,
        train_loader,
        loss_fn,
        config["num_epochs"],
        config["lr"],
        config["weight_decay"],
        config["epoch_lr_restart"],
        val_loader,
        report_fn,
        resume_states,
        device,
    )
    if model_file is not None:
        save_model(model, model_file)


def init_mnist_model(
    config: dict[str, int | float | list[int]], device: torch.device = "cpu"
) -> torch.nn.Module:
    """Initializes a vision transformer for training on MNIST.

    Args:
        config (dict): Training configuration including `'patch_size'`, `'num_heads'`,
            `'latent_size_multiplier'`, `'num_layers'`, `'encoder_size'`, `'head_size'`
            and `'dropout'`.
        device (torch.device, optional): Device to load the model on.
            Default: `'cpu'`.
    Returns:
        torch.nn.Module: The initialized model.
    """
    model = VisionTransformer(
        num_channels=1,
        input_sizes=[28, 28],
        output_size=10,
        patch_size=config["patch_size"],
        num_heads=config["num_heads"],
        latent_size_multiplier=config["latent_size_multiplier"],
        num_layers=config["num_layers"],
        encoder_size=config["encoder_size"],
        head_size=config["head_size"],
        dropout=config["dropout"],
    )
    model = model.to(device)
    return model


def train(
    model: torch.nn.Module,
    train_loader: torch.utils.data.DataLoader,
    loss_fn: torch.nn.Module,
    num_epochs: int,
    lr: float,
    weight_decay: float,
    epoch_lr_restart: int,
    val_loader: torch.utils.data.DataLoader = None,
    report_fn: Callable[
        [
            int,
            float,
            float,
            torch.nn.Module,
            torch.optim.Optimizer,
            torch.optim.lr_scheduler.LRScheduler,
        ],
        None,
    ] = None,
    resume_states: dict[str, int | dict[str, Any]] = None,
    device: torch.device = "cpu",
) -> None:
    """Main training function for model training.

    Initializes an optimizer and learning rate scheduler, contains the loop over epochs
    and eventually evaluates validation performance.

    Args:
        model (torch.nn.Module): Model to train.
        train_loader (torch.utils.data.DataLoader): Training data loader.
        loss_fn (torch.nn.Module): Loss function for model training.
        num_epochs (int): The number of epochs to use.
        lr (float): Learning rate.
        weight_decay (float): Weight decay coefficient.
        epoch_lr_restart (int): Epoch for first learning rate scheduler restart.
        val_loader (torch.utils.data.DataLoader, optional): Validation data loader.
            Default: `None`.
        report_fn (callable, optional): A function for reporting the training state.
            The function must accept arguments for epoch number (`int`),
            training loss (`float`), validation loss (`float`),
            model (`torch.nn.Module`),
            optimizer (`torch.optim.Optimizer`) and
            lr_scheduler (`torch.optim.lr_scheduler.LRScheduler`).  Default: `None`.
        resume_states (dict, optional): Dictionary with states for `'epoch'`,
            `'optimizer'` and `'lr_scheduler'`.  Default: `None`.
        device (torch.device, optional): Device to train the model on.
            Default: `'cpu'`.
    """
    optimizer = torch.optim.AdamW(model.parameters(), lr=lr, weight_decay=weight_decay)
    lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(
        optimizer=optimizer,
        T_0=epoch_lr_restart,
    )
    if resume_states is None:
        start_epoch = 0
    else:
        start_epoch = resume_states["epoch"] + 1
        optimizer.load_state_dict(resume_states["optimizer"])
        lr_scheduler.load_state_dict(resume_states["lr_scheduler"])
    iters = len(train_loader)
    for epoch in range(start_epoch, num_epochs):
        model.train()
        for step, (data, target) in enumerate(train_loader):
            data, target = data.to(device), target.to(device)
            optimizer.zero_grad()
            output = model(data)
            loss = loss_fn(output, target)
            loss.backward()
            optimizer.step()
            lr_scheduler.step(epoch + step / iters)
        if val_loader is None:
            val_loss = None
        else:
            val_loss = prediction_loss(model, val_loader, loss_fn, device)
        if report_fn is not None:
            report_fn(epoch, loss, val_loss, model, optimizer, lr_scheduler)


def main() -> None:
    """Processes command line arguments with training."""
    parser = argparse.ArgumentParser(description="MNIST Vision Transformer Training")
    parser.add_argument(
        "--batch-size",
        type=int,
        default=32,
        metavar="N",
        help="input batch size for training (default: 32)",
    )
    parser.add_argument(
        "--num-epochs",
        type=int,
        default=60,
        metavar="N",
        help="number of epochs to train (default: 60)",
    )
    parser.add_argument(
        "--lr",
        type=float,
        default=1e-4,
        metavar="R",
        help="learning rate (default: 1e-4)",
    )
    parser.add_argument(
        "--weight-decay",
        type=float,
        default=2e-2,
        metavar="R",
        help="weight decay coefficient (default: 2e-2)",
    )
    parser.add_argument(
        "--epoch-lr-restart",
        type=int,
        default=22,
        metavar="N",
        help="epoch for first scheduler restart (default: 22)",
    )
    parser.add_argument(
        "--patch-size",
        type=int,
        default=4,
        metavar="P",
        help="single dimension size of an image patch (default: 4)",
    )
    parser.add_argument(
        "--num-heads",
        type=int,
        default=28,
        metavar="N",
        help="number of attention heads (default: 28)",
    )
    parser.add_argument(
        "--latent-size-multiplier",
        type=int,
        default=15,
        metavar="M",
        help="yields latent size when multiplied with num_heads (default: 15)",
    )
    parser.add_argument(
        "--num-layers",
        type=int,
        default=9,
        metavar="L",
        help="number of encoder blocks (default: 9)",
    )
    parser.add_argument(
        "--encoder-size",
        type=int,
        default=615,
        metavar="H",
        help="number of hidden units of transformer encoder MLPs (default: 615)",
    )
    parser.add_argument(
        "--head-size",
        type=int,
        default=88,
        metavar="H",
        help="number of hidden units of MLP head (default: 88)",
    )
    parser.add_argument(
        "--dropout",
        type=float,
        default=6e-2,
        metavar="R",
        help="dropout rate (default: 6e-2)",
    )
    parser.add_argument(
        "--model-file",
        type=str,
        default="model.pt",
        metavar="FILE",
        help="file to save the model to (default: 'model.pt')",
    )
    parser.add_argument(
        "--use-validation",
        action="store_true",
        default=False,
        help="enables validation set",
    )
    parser.add_argument(
        "--no-augmentation",
        action="store_true",
        default=False,
        help="disables data augmentation",
    )
    parser.add_argument(
        "--seed", type=int, default=1, metavar="S", help="random seed (default: 1)"
    )
    parser.add_argument(
        "--no-cuda", action="store_true", default=False, help="disables CUDA training"
    )
    args = parser.parse_args()
    no_cuda = args.no_cuda or not torch.cuda.is_available()
    device = torch.device("cpu" if no_cuda else "cuda")
    torch.manual_seed(args.seed)
    config = {
        "batch_size": args.batch_size,
        "num_epochs": args.num_epochs,
        "lr": args.lr,
        "weight_decay": args.weight_decay,
        "epoch_lr_restart": args.epoch_lr_restart,
        "patch_size": args.patch_size,
        "num_heads": args.num_heads,
        "latent_size_multiplier": args.latent_size_multiplier,
        "num_layers": args.num_layers,
        "encoder_size": args.encoder_size,
        "head_size": args.head_size,
        "dropout": args.dropout,
    }

    # Define function for reporting training progress
    def report_fn(
        epoch: int,
        train_loss: float,
        val_loss: float,
        model: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        lr_scheduler: torch.optim.lr_scheduler.LRScheduler,
    ) -> None:
        print(
            f"epoch: {epoch+1}\t"
            f"training loss: {train_loss}\t"
            f"validation loss: {val_loss}"
        )

    use_augmentation = not args.no_augmentation
    train_mnist(
        config,
        data_dir="data",
        use_validation=args.use_validation,
        use_augmentation=use_augmentation,
        model_file=args.model_file,
        report_fn=report_fn,
        device=device,
    )


if __name__ == "__main__":
    main()



from os import PathLike
from typing import IO, BinaryIO, Union

import torch
from typing_extensions import TypeAlias

from .model import VisionTransformer

FILE_LIKE: TypeAlias = Union[str, PathLike, BinaryIO, IO[bytes]]


def save_model(model: VisionTransformer, model_file: FILE_LIKE) -> None:
    """Saves the vision transformer model, including model keyword arguments, to a file.

    Args:
        model (mnistvit.model.VisionTransformer): The model to save.
        model_file (FILE_LIKE): File name to save the model to.
    """
    model_dict = {
        "kwargs": model.kwargs,
        "state_dict": model.state_dict(),
    }
    torch.save(model_dict, model_file)


def load_model(
    model_file: FILE_LIKE, device: torch.device = "cpu"
) -> VisionTransformer:
    """Loads the model from a file.

    Args:
        model_file (FILE_LIKE): File name to load the model from.
        device (torch.device, optional): Device to load the model to.  Default: `'cpu'`.

    Returns:
        mnistvit.model.VisionTransformer: The loaded model.
    """
    model_dict = torch.load(model_file, map_location=torch.device("cpu"))
    model = VisionTransformer(**model_dict["kwargs"])
    model.load_state_dict(model_dict["state_dict"])
    model = model.to(device)
    return model



import argparse
import os
from tempfile import TemporaryDirectory

import torch
from ray import train, tune
from ray.tune.schedulers import ASHAScheduler
from ray.tune.search.optuna import OptunaSearch

from .train import init_mnist_model, train_mnist
from .utils import FILE_LIKE, save_model


def objective(config: dict[str, int | float], data_dir: str) -> None:
    """Objective function for hyperparameter tuning.

    Trains a model on MNIST according to the configuration and reports the mean loss.
    Also saves checkpoints to `'checkpoint.pt'` files.  Checkpoints contain model,
    optimizer and learning rate scheduler states as well as `'epoch'` metadata.

    Args:
        config (dict): Training configuration including `'batch_size'`, `'num_epochs'`,
            `'lr'`, `'weight_decay'`, `'epoch_lr_restart'`, `'patch_size'`,
            `'num_heads'`, `'latent_size_multiplier'`, `'num_layers'`, `'encoder_size'`,
            `'head_size'` and `'dropout'`.
        data_dir (str): Directory of the MNIST training data.
    """

    # Define callback function for checkpoint saving
    def report_fn(
        epoch: int,
        train_loss: float,
        val_loss: float,
        model: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        lr_scheduler: torch.optim.lr_scheduler.LRScheduler,
    ) -> None:
        metrics = {"mean_loss": val_loss}
        with TemporaryDirectory() as temp_dir:
            torch.save(
                (model.state_dict(), optimizer.state_dict(), lr_scheduler.state_dict()),
                os.path.join(temp_dir, "checkpoint.pt"),
            )
            metadata = {"epoch": epoch}
            checkpoint = train.Checkpoint.from_directory(temp_dir)
            checkpoint.set_metadata(metadata)
            train.report(metrics=metrics, checkpoint=checkpoint)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # Resume checkpoint if available
    if train.get_checkpoint():
        checkpoint = train.get_checkpoint()
        with checkpoint.as_directory() as checkpoint_dir:
            model_state, optimizer_state, lr_scheduler_state = torch.load(
                os.path.join(checkpoint_dir, "checkpoint.pt")
            )
            resume_states = {
                "epoch": checkpoint.get_metadata()["epoch"],
                "model": model_state,
                "optimizer": optimizer_state,
                "lr_scheduler": lr_scheduler_state,
            }
    else:
        resume_states = None

    train_mnist(
        config=config,
        data_dir=data_dir,
        report_fn=report_fn,
        resume_states=resume_states,
        device=device,
    )


def fit(
    exp_name: str,
    storage_path: str,
    num_samples: int,
    num_epochs: int,
    model_file: FILE_LIKE,
    resources: dict[str, float] = None,
) -> None:
    """Tunes hyperparameters of a model to MNIST.

    Selects the checkpoint with the best validation performance and prints the best
    result and the best checkpoint metadata.  The best model is then saved to the
    provided model file name.

    Args:
        exp_name (str): Name of the experiment.
        storage_path (str): Path of the experiment directory.
        num_samples (int): The number of hyperparameter configurations to try.
        num_epochs (int): The number of epochs per optimization.
        model_file (FILE_LIKE): File name to save the best model to.
        resources (dict, optional): Resource configuration per trial.  Default: `None`.
    """
    search_space = {
        "batch_size": tune.choice([32, 64, 128, 256]),
        "num_epochs": num_epochs,
        "lr": tune.loguniform(1e-5, 0.01),
        "weight_decay": tune.loguniform(1e-4, 0.1),
        "epoch_lr_restart": tune.choice([4, 8, 16, 32, 64]),
        "patch_size": tune.choice([2, 4, 7, 14]),
        "num_heads": tune.choice([2, 4, 8, 16]),
        "latent_size_multiplier": tune.choice([4, 8, 16, 32]),
        "num_layers": tune.choice([1, 2, 4, 8]),
        "encoder_size": tune.choice([2**i for i in range(4, 10)]),
        "head_size": tune.choice([2**i for i in range(4, 10)]),
        "dropout": tune.uniform(0, 0.5),
    }
    data_dir = os.path.abspath("data")
    trainable = tune.with_parameters(objective, data_dir=data_dir)
    metric, mode = "mean_loss", "min"
    if resources is not None:
        trainable = tune.with_resources(trainable, resources=resources)
    storage_path = os.path.abspath(storage_path)
    exp_path = os.path.join(storage_path, exp_name)
    if tune.Tuner.can_restore(exp_path):
        tuner = tune.Tuner.restore(
            exp_path,
            trainable=trainable,
            resume_errored=True,
            restart_errored=False,
            resume_unfinished=True,
            param_space=search_space,
        )
    else:
        tuner = tune.Tuner(
            trainable,
            run_config=train.RunConfig(
                name=exp_name,
                storage_path=storage_path,
                checkpoint_config=train.CheckpointConfig(
                    checkpoint_score_attribute=metric,
                    checkpoint_score_order=mode,
                    num_to_keep=5,
                ),
            ),
            tune_config=tune.TuneConfig(
                num_samples=num_samples,
                search_alg=OptunaSearch(),
                scheduler=ASHAScheduler(),
                metric=metric,
                mode=mode,
            ),
            param_space=search_space,
        )
    results = tuner.fit()
    best_result = results.get_best_result(scope="all")
    best_checkpoint = best_result.get_best_checkpoint(
        metric=metric,
        mode=mode,
    )
    print("Best result config: ", best_result.config)
    print("Best checkpoint: ", best_checkpoint.get_metadata())
    with best_checkpoint.as_directory() as checkpoint_dir:
        model_state = torch.load(
            os.path.join(checkpoint_dir, "checkpoint.pt"),
            map_location=torch.device("cpu"),
        )[0]
    model = init_mnist_model(best_result.config)
    model.load_state_dict(model_state)
    save_model(model, model_file)


def main() -> None:
    """Processes command line arguments with tuning."""
    parser = argparse.ArgumentParser(description="MNIST Vision Transformer Tuning")
    parser.add_argument(
        "--exp-name",
        type=str,
        default="mnistvit",
        metavar="NAME",
        help="name of the experiment to run (default: 'mnistvit')",
    )
    parser.add_argument(
        "--storage-path",
        type=str,
        default="ray_results",
        metavar="DIR",
        help="path of the experiment directory (default: 'ray_results')",
    )
    parser.add_argument(
        "--num-samples",
        type=int,
        default=1024,
        metavar="N",
        help="number of configs to test (default: 1024)",
    )
    parser.add_argument(
        "--num-epochs",
        type=int,
        default=64,
        metavar="N",
        help="number of epochs to train (default: 64)",
    )
    parser.add_argument(
        "--model-file",
        type=str,
        default="model.pt",
        metavar="FILE",
        help="file to save the best model to (default: 'model.pt')",
    )
    parser.add_argument(
        "--cpu-resource",
        type=float,
        default=None,
        metavar="R",
        help="CPU resource per trial (default: None)",
    )
    parser.add_argument(
        "--gpu-resource",
        type=float,
        default=None,
        metavar="R",
        help="GPU resource per trial (default: None)",
    )
    args = parser.parse_args()
    if args.cpu_resource is None and args.gpu_resource is None:
        resources = None
    else:
        resources = {}
        if args.cpu_resource is not None:
            resources["cpu"] = args.cpu_resource
        if args.gpu_resource is not None:
            resources["gpu"] = args.gpu_resource
    fit(
        exp_name=args.exp_name,
        storage_path=args.storage_path,
        num_samples=args.num_samples,
        num_epochs=args.num_epochs,
        model_file=args.model_file,
        resources=resources,
    )


if __name__ == "__main__":
    main()



import torch
from torch.utils.data import DataLoader, random_split
from torchvision import datasets, transforms
from torchvision.io import ImageReadMode, read_image

from .utils import FILE_LIKE


def preprocess_mnist(
    data_dir: str, train: bool, use_augmentation: bool = False
) -> torch.utils.data.dataset.Dataset:
    """Preprocesses the MNIST dataset.

    Normalizes the MNIST dataset with training mean and variance and eventually
    augments the dataset with random affine transformations.

    Args:
        data_dir (str): Directory of the MNIST dataset.
        train (bool): If true, loads the training set, else the test set.
        use_augmentation (bool, optional): If true, augments the dataset with random
            affine transformations.  Default: `False`.

    Returns:
        torch.utils.data.dataset.Dataset: Image, target pairs.
    """
    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )
    if use_augmentation:
        # Data augmentation
        transform = transforms.Compose(
            [
                transforms.RandomAffine(
                    degrees=16, translate=(0.1, 0.1), scale=(0.9, 1.1)
                ),
                transform,
            ]
        )
    dataset = datasets.MNIST(data_dir, train=train, download=True, transform=transform)
    return dataset


def get_train_loaders_mnist(
    data_dir: str,
    batch_size: int,
    train_fraction: float = 1.0,
    use_augmentation: bool = True,
) -> tuple[DataLoader, DataLoader]:
    """Training loaders of the MNIST dataset.

    Args:
        data_dir (str): Directory of the MNIST dataset.
        batch_size (int): Size of the batches of the training loaders.
        train_fraction (float, optional): Fraction of the set used for the training
            loader.  The remainder is used for the validation loader.  Default: 1.0.
        use_augmentation (bool, optional): If true, augments the dataset with random
            affine transformations.  Default: `True`.

    Returns:
        tuple: Training loader and validation loader, where the latter is `None` if
            `train_fraction` is 1.
    """
    assert 0 <= train_fraction <= 1
    loader_kwargs = {"shuffle": True, "batch_size": batch_size}
    dataset = preprocess_mnist(data_dir, train=True, use_augmentation=use_augmentation)
    if train_fraction == 1.0:
        train_loader = DataLoader(dataset, **loader_kwargs)
        val_loader = None
    else:
        num_train = int(len(dataset) * train_fraction)
        train_set, val_set = random_split(
            dataset, [num_train, len(dataset) - num_train]
        )
        train_loader = DataLoader(train_set, **loader_kwargs)
        val_loader = DataLoader(val_set, **loader_kwargs)
    return train_loader, val_loader


def get_test_loader_mnist(data_dir: str, batch_size: int) -> DataLoader:
    """Test loader of the MNIST dataset.

    Args:
        data_dir (str): Directory of the MNIST dataset.
        batch_size (int): Size of the batches of the test loader.

    Returns:
        torch.utils.data.DataLoader: Test loader.
    """
    dataset = preprocess_mnist(data_dir, train=False)
    loader = DataLoader(dataset, shuffle=False, batch_size=batch_size)
    return loader


def read_digit_image(image_file: FILE_LIKE) -> torch.FloatTensor:
    """Loads a single digit image from a file.

    Center crops and resizes the image to 28 by 28 pixels.  Also inverts the image if
    there are more bright than dark pixels.

    Args:
        image_file (FILE_LIKE): The image file.

    Returns:
        torch.FloatTensor: The preprocessed digit image.
    """
    image = read_image(image_file, mode=ImageReadMode.GRAY)
    transform = transforms.Compose(
        [
            transforms.CenterCrop(min(image.size()[1:])),
            transforms.Resize([28, 28]),
        ]
    )
    # Standardize image
    image = transform(image).type(torch.FloatTensor)
    image = transforms.Normalize(image.mean(), image.std())(image)
    # Check if we need to invert the image
    if (image > 0).count_nonzero() > image.numel() / 2:
        # More bright than dark pixels
        image = transforms.functional.invert(image)
    return image
